"""
Lazy Evaluation for Python - main package with primary exports

Copyright (c) 2004, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import sys
assert sys.hexversion >= 0x02030000, 'at least Python 2.3 is needed'

from Promises import Promise, PromiseMetaClass, force
from Futures import Future
from ForkedFutures import ForkedFuture
from LazyClasses import LazyEvaluated, LazyEvaluatedMetaClass
from Functions import delay, lazy, spawn, future, fork, forked




"""
Lazy Evaluation for Python - utility functions

Copyright (c) 2004, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

class NoneSoFar:

    """
    This is a singleton to give you something to put somewhere that
    should never be a rightfull return value of anything.
    """
    
    def __str__(self):
        return 'NoneSoFar'
    
    def __repr__(self):
        return 'NoneSoFar'

    def __nonzero__(self):
        return 0

NoneSoFar = NoneSoFar()

def getitem(obj, key):
    """
    This is a helper function needed in promise objects to pass
    on __getitem__ calls. It just mimicks the getattr call, only
    it uses dictionary style access.
    """
    return obj[key]

def setitem(obj, key, value):
    """
    This is a helper function needed in promise objects to pass
    on __setitem__ calls. It just mimicks the setattr call, only
    it uses dictionary style access.
    """
    obj[key] = value

def delitem(obj, key):
    """
    This is a helper function needed in promise objects to pass
    on __delitem__ calls. It just mimicks the delattr call, only
    it uses dictionary style access.
    """
    del obj[key]

def getslice(obj, start, stop):
    """
    This is a helper function needed in promise objects to pass
    on __getslice__ calls. It just mimicks the getattr call, only
    it uses dictionary style access.
    """
    return obj[start:stop]

def setslice(obj, start, stop, value):
    """
    This is a helper function needed in promise objects to pass
    on __setslice__ calls. It just mimicks the setattr call, only
    it uses dictionary style access.
    """
    obj[start:stop] = value

def delslice(obj, start, stop):
    """
    This is a helper function needed in promise objects to pass
    on __delslice__ calls. It just mimicks the delattr call, only
    it uses dictionary style access.
    """
    del obj[start:stop]




"""
Lazy Evaluation for Python - functions for lazy function calling

Copyright (c) 2004, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from Promises import Promise
from Futures import Future
from ForkedFutures import ForkedFuture

def delay(funk, args=None, kw=None, promiseclass=Promise):

    """
    This is a lazy variant on the apply function. It returns a promise
    for the function call that will be evaluated when needed. You can
    override the class to be used for the promise.
    """

    if args is None: args = []
    if kw is None: kw = {}
    return promiseclass(funk, args, kw)

def lazy(funk, promiseclass=Promise):

    """
    This function returns a lazy variant on the passed in function.
    That lazy variant will not directly evaluate but will push that
    evaluation off to some promise point. The class to be used for
    the promise can be overridden.
    """

    def lazy_funk(*args, **kw):
        return promiseclass(funk, args, kw)

    return lazy_funk

def spawn(funk, args=None, kw=None, futureclass=Future):

    """
    This is a parallel variant on the apply function. It returns a future
    for the function call that will be evaluated in the background. You can
    override the class to be used for the future.
    """

    if args is None: args = []
    if kw is None: kw = {}
    return futureclass(funk, args, kw)

def future(funk, futureclass=Future):

    """
    This function returns a future variant on the passed in function.
    That future variant will not directly evaluate but will push that
    evaluation off to the background. The class to be used for
    the future can be overridden. Every call to a future function will
    return a new future.
    """

    def future_funk(*args, **kw):
        return futureclass(funk, args, kw)

    return future_funk

def fork(funk, args=None, kw=None, futureclass=ForkedFuture):

    """
    This is a parallel variant on the apply function. It returns a future
    for the function call that will be evaluated in the background. You can
    override the class to be used for the future. It uses forked futures
    by default.
    """

    if args is None: args = []
    if kw is None: kw = {}
    return futureclass(funk, args, kw)

def forked(funk, futureclass=ForkedFuture):

    """
    This function returns a future variant on the passed in function.
    That future variant will not directly evaluate but will push that
    evaluation off to the background. The class to be used for
    the future can be overridden. Every call to a future function will
    return a new future. It uses forked futures by default.
    """

    def future_funk(*args, **kw):
        return futureclass(funk, args, kw)

    return future_funk




"""
Lazy Evaluation for Python - convenience classes

Copyright (c) 2004, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import types

from Promises import Promise
from Functions import lazy

class LazyEvaluatedMetaClass(type):

    """
    This meta class rewrites all function attributes to not directly
    run but to yield a generator that will run the function later on.
    """

    def __init__(klass, name, bases, attributes):
        promiseclass = getattr(klass, '__promiseclass__', Promise)
        for (k, v) in attributes.items():
            if isinstance(v, types.FunctionType):
                setattr(klass, k, lazy(v, promiseclass))
        super(LazyEvaluatedMetaClass, klass).__init__(name,
            bases, attributes)

class LazyEvaluated(object):

    """
    This is the base class for all classes that should evaluate in
    a lazy fashion. You can overload __promiseclass__ if you want to
    have different promise handling in your code.
    """

    __metaclass__ = LazyEvaluatedMetaClass
    __promiseclass__ = Promise




"""
Lazy Evaluation for Python - forked futures (high level concurrency) for python

Copyright (c) 2010, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from multiprocessing import Process, Queue
from Promises import Promise, PromiseMetaClass
from Utils import NoneSoFar

class ForkedFuture(object):

    """
    This class builds future objects. A future is something that
    will be evaluated in parallel to the current flow of command.
    If you access the value of a future, your code will block until
    the value becomes available.

    The initialization get's the function and it's parameters to
    delay. If this is a promise that is created because of a delayed
    method on a promise, args[0] will be another promise of the same
    class as the current promise and func will be one of (getattr,
    apply, getitem, getslice). This knowledge can be used to optimize
    chains of delayed functions. Method access on promises will be
    factored as one getattr promise followed by one apply promise.

    A delayed future (applying lazy operators - either the lazy HOF
    or some of getattr, getitem, getslice or call) will create a
    normal promise.
    """

    __metaclass__ = PromiseMetaClass
    __delayclass__ = Promise

    def __init__(self, func, args, kw):
        """
        Start a process with the function to be computed. Block the
        result with a queue so that somebody trying to force the
        value will block until we are complete. If the process
        get's an exception, store that for raising on force.
        """

        def thunk():
            try:
                res = apply(func, args, kw)
                self.__queue.put((True, res))
            except Exception, e:
                self.__queue.put((False, e))

        self.__queue = Queue()
        self.__result = NoneSoFar
        self.__exception = NoneSoFar
        self.__proc = Process(target=thunk)
        self.__proc.start()
    
    def __force__(self):
        """
        This function returns either the value or the exception
        of the future. If the future hasn't completed yet, this
        call will block until it has.
        """

        if self.__result is NoneSoFar and self.__exception is NoneSoFar:
            (f, v) = self.__queue.get()
            if f:
                self.__result = v
            else:
                self.__exception = v
        if self.__result is not NoneSoFar:
            return self.__result
        elif self.__exception is not NoneSoFar:
            raise self.__exception
    



from setuptools import setup

setup(
    version = "0.3",
    description = "Lazy Evaluation for Python",
    author = "Georg Bauer",
    author_email = "gb@rfc1437.de",
    url = "http://bitbucket.org/rfc1437/lazypy/",
    name='lazypy', 
    long_description=file("README").read(),
    license='MIT/X',
    platforms=['BSD','Linux','MacOS X', 'win32'],
    packages=['LazyEvaluation'],
    scripts=['test.py'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
)



import sys
import unittest

from LazyEvaluation import *
from LazyEvaluation.Utils import NoneSoFar

class MySpecialError(Exception):
    pass

def anton(a,b):
    return a+b

class ClassWithAttrs:
    pass

class ClassWithLazyMethod:

    def __init__(self):
        self.attr = self.anton(5,6)

    def anton(self, a, b):
        return a+b
    anton = lazy(anton)

class MyPromise(object):

    __metaclass__ = PromiseMetaClass

    def __init__(self, func, args, kw):
        self.__func = func
        self.__args = args
        self.__kw = kw
        self.__result = NoneSoFar

    def forced(self):
        return self.__result is not NoneSoFar
    
    def __force__(self):
        if self.__result is NoneSoFar:
            args = [force(arg) for arg in self.__args]
            kw = dict([(k, force(v)) for (k, v)
                    in self.__kw.items()])
            self.__result = apply(self.__func, args, kw)
        return self.__result

class LazyClass(LazyEvaluated):

    __promiseclass__ = MyPromise

    def anton(self, a, b):
        return a+b

    def berta(self, a, b):
        return a*b

    def caesar(self):
        return 'blah'
    
    def detlef(self):
        res = ClassWithAttrs()
        res.blah = ClassWithAttrs()
        res.blah.blubb = 5
        return res

class TestCase100Simple(unittest.TestCase):

    def testDelay(self):
        promise = delay(anton, (5, 6))
        self.assertTrue(isinstance(promise, Promise))
        self.assertEqual(promise, 11)
        self.assertEqual(promise, 11)
    
    def testDelayList(self):

        def berta(a,b):
            return range(a,b)

        promise = delay(berta, (0, 10))
        self.assertTrue(isinstance(promise, Promise))
        self.assertEqual(len(promise), 10)
        self.assertEqual(len(promise), 10)
    
    def testIntegers(self):
        funk = lazy(anton)
        self.assertTrue(isinstance(funk(5,6), Promise))
        self.assertEqual(funk(5,6), 11)
        self.assertEqual(str(funk(5,6)), '11')
        self.assertEqual(-funk(5,6), -11)

    def testBinary(self):
        funk = lazy(anton)
        self.assertEqual(funk(5,6)|funk(9,7), 27)
        self.assertEqual(funk(5,6)&funk(9,7), 0)
    
    def testIntegerCombined(self):
        funk = lazy(anton)
        self.assertEqual(funk(5,6)+11, 22)
        self.assertEqual(11+funk(5,6), 22)
        self.assertEqual(funk(5,6)*3, 33)
        self.assertEqual(3*funk(5,6), 33)
    
    def testBools(self):
        funk = lazy(anton)
        self.assertTrue(funk(5,6))
        self.assertFalse(funk(5,-5))

    def testFloats(self):
        funk = lazy(anton)
        self.assertTrue(isinstance(funk(5.1,6.2), Promise))
        self.assertEqual(funk(5.1,6.2), 11.3)

    def testLongs(self):
        funk = lazy(anton)
        self.assertTrue(isinstance(funk(5L,6L), Promise))
        self.assertEqual(funk(3333333333333L,5555555555555L), 8888888888888L)

    def testStrings(self):
        funk = lazy(anton)
        self.assertTrue(isinstance(funk('anton','berta'), Promise))
        self.assertEqual(str(funk('anton', 'berta')), 'antonberta')
        self.assertEqual(funk('anton', 'berta'), 'antonberta')
        self.assertTrue(funk('anton', 'berta') > 'anton')

    def testStringConcat(self):
        funk = lazy(anton)
        self.assertEqual(funk('anton', 'berta') + 'blah', 'antonbertablah')
        self.assertEqual('blah' + funk('anton', 'berta'), 'blahantonberta')

    def testStringMult(self):
        funk = lazy(anton)
        self.assertEqual(funk('anton','berta')*2, 'antonbertaantonberta')
        self.assertEqual(2*funk('anton','berta'), 'antonbertaantonberta')
        self.assertEqual('blah'*funk(1,1), 'blahblah')
        self.assertEqual(funk(1,1)*'blah', 'blahblah')
    
    def testUnicode(self):
        funk = lazy(anton)
        self.assertTrue(isinstance(funk(u'anton',u'berta'), Promise))
        self.assertEqual(unicode(funk('anton', 'berta')), u'antonberta')
        self.assertEqual(funk(u'anton', u'berta'), u'antonberta')
    
    def testAttribute(self):
        funk = lazy(anton)
        a = getattr(funk(5,6), 'isnich', 99)
        self.assertEqual(a, 99)
    
    def testStringInterpolation(self):
        funk = lazy(anton)
        self.assertEqual((funk('blah%s', 'blubb%d') % ('anton', 5)), 'blahantonblubb5')

class TestCase200LazyDicts(unittest.TestCase):

    def setUp(self):
        def anton(a,b):
            return dict([(x,x) for x in range(a,b)])
        
        self.g = lazy(anton)

    def testCreate(self):
        hash = self.g(1,6)
        self.assertEqual(hash, {1:1, 2:2, 3:3, 4:4, 5:5})
    
    def testKeys(self):
        hash = self.g(1,6)
        l = force(hash).keys()
        l.sort()
        self.assertEqual(l, [1,2,3,4,5])

    def testValues(self):
        hash = self.g(1,6)
        l = force(hash).values()
        l.sort()
        self.assertEqual(l, [1,2,3,4,5])
    
    def testAccess(self):
        hash = self.g(1,6)
        self.assertEqual(hash[3], 3)
    
class TestCase300LazyClass(unittest.TestCase):

    def setUp(self):
        self.obj = LazyClass()
    
    def testPromises(self):
        self.assertTrue(isinstance(self.obj.anton(5,6), MyPromise))
    
    def testForcing(self):
        promise = self.obj.anton(5,6)
        self.assertTrue(isinstance(promise, MyPromise))
        self.assertFalse(promise.forced())
        self.assertEqual(promise, 11)
        self.assertTrue(promise.forced())

    def testDirectAccess(self):
        self.assertEqual(self.obj.anton(5,6), 11)

    def testBinaryOperator(self):
        self.assertEqual(self.obj.anton(5,6)+self.obj.berta(5,6), 41)

    def testMixedNumeric(self):
        self.assertEqual(self.obj.anton(5,6)+5, 16)
        self.assertEqual(5+self.obj.anton(5,6), 16)

    def testStringMultiply(self):
        self.assertEqual(self.obj.caesar()*3, 'blahblahblah')

    def testAttributeAccess(self):
        self.assertEqual(force(self.obj.detlef()).blah.blubb, 5)

    def testInlineModification(self):
        anton = self.obj.anton(5,6)
        anton += 11
        self.assertEqual(anton, 22)

class TestCase400LazyLists(unittest.TestCase):

    def setUp(self):
        
        def berta(a,b):
            return range(a,b)

        self.func = lazy(berta)

    def testCreate(self):
        l = self.func(0,10)
        self.assertTrue(isinstance(l, Promise))
        self.assertEqual(l, range(0,10))
    
    def testAccess(self):
        self.assertEqual(self.func(0,10)[6], 6)
        self.assertEqual(self.func(0,10)[0], 0)
        self.assertEqual(self.func(0,10)[-1], 9)
    
    def testSlice(self):
        self.assertEqual(self.func(0,10)[6:8], [6,7])
        self.assertEqual(self.func(0,10)[:2], [0,1])
        self.assertEqual(self.func(0,10)[-2:], [8,9])
        self.assertEqual(self.func(0,10)[:], range(0,10))

    def testExtendedSlice(self):
        self.assertEqual(self.func(0,10)[1:5:2], [1,3])
    
    def testLen(self):
        self.assertEqual(len(self.func(0,5)), 5)
    
    def testIter(self):
        l = []
        for el in self.func(0,10):
            l.append(el)
        self.assertEqual(l, range(0,10))
    
    def testListComprehension(self):
        l = [int(x) for x in self.func(0,10)]
        self.assertEqual(l, range(0,10))

class TestCase500Futures(unittest.TestCase):

    def testFastFuture(self):
        f = spawn(lambda : 5+6)
        self.assertTrue(isinstance(f, Future))
        self.assertEqual(f, 11)

    def testLongerFuture(self):
        
        def fib(n):
            if n in (0,1):
                return 1
            return fib(n-1) + fib(n-2)

        f = future(fib)
        for n in (5, 10, 20):
            self.assertEqual(f(n), fib(n))
    
class TestCase550ForkedFutures(unittest.TestCase):

    def testFastFuture(self):
        f = fork(lambda : 5+6)
        self.assertTrue(isinstance(f, ForkedFuture))
        self.assertEqual(f, 11)
        self.assertEqual(f, 11)
        self.assertEqual(f, 11)

    def testFutureWithException(self):
        def crasher():
            raise MySpecialError(55)

        f = fork(crasher)
        self.assertTrue(isinstance(f, ForkedFuture))
        self.assertRaises(MySpecialError, crasher)

    def testLongerFuture(self):
        
        def fib(n):
            if n in (0,1):
                return 1
            return fib(n-1) + fib(n-2)

        f = forked(fib)
        for n in (5, 10, 20, 30):
            self.assertEqual(f(n), fib(n))
    
class TestCase600LazyMethod(unittest.TestCase):

    def testAttribute(self):
        o = ClassWithLazyMethod()
        self.assertTrue(isinstance(o.attr, Promise))
        self.assertEqual(o.attr, 11)
    
    def testNonZero(self):
        o = ClassWithLazyMethod()
        self.assertTrue(isinstance(o.attr, Promise))
        self.assertEqual(getattr(o.attr, 'blah', True), True)

    def testString(self):
        o = ClassWithLazyMethod()
        self.assertTrue(isinstance(o.attr, Promise))
        self.assertEqual(('a%sb' % o.attr), 'a11b')

    def testInteger(self):
        o = ClassWithLazyMethod()
        self.assertTrue(isinstance(o.attr, Promise))
        self.assertEqual(5+o.attr, 16)

if __name__ == '__main__':
    unittest.main()




"""
Lazy Evaluation for Python - promises for python

Copyright (c) 2004, Georg Bauer <gb@rfc1437.de>, except where the file
explicitly names other copyright holders and licenses.

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from Utils import *

def force(value):
    """
    This helper function forces evaluation of a promise. A promise
    for this function is something that has a __force__ method (much
    like an iterator in python is anything that has a __iter__
    method).
    """

    f = getattr(value, '__force__', None)
    if f: return f()
    else: return value

class PromiseMetaClass(type):

    """
    This meta class builds the behaviour of promise classes. It's mainly
    building standard methods with special behaviour to mimick several
    types in Python.
    
    The __magicmethods__ list defines what magic methods are created. Only
    those magic methods are defined that are not already defined by the
    class itself.
    
    __magicrmethods__ is much like __magicmethods__ only that it provides
    both the rmethod and the method so the proxy can decide what to use.
    
    The __magicfunctions__ list defines methods that should be mimicked by
    using some predefined function.
    
    The promise must define a __force__ method that will force evaluation
    of the promise.
    """

    __magicmethods__ = ['__abs__', '__pos__', '__invert__', '__neg__']
    
    __magicrmethods__ = [('__radd__', '__add__'), ('__rsub__', '__sub__'),
        ('__rdiv__', '__div__'), ('__rmul__', '__mul__'),
        ('__rand__', '__and__'), ('__ror__', '__or__'),
        ('__rxor__', '__xor__'), ('__rlshift__', '__lshift__'),
        ('__rrshift__', '__rshift__'), ('__rmod__', '__mod__'),
        ('__rdivmod__', '__divmod__'), ('__rtruediv__', '__truediv__'),
        ('__rfloordiv__', '__floordiv__'), ('__rpow__', '__pow__')]
    
    __magicfunctions__ = [('__cmp__', cmp), ('__str__', str),
        ('__unicode__', unicode), ('__complex__', complex),
        ('__int__', int), ('__long__', long), ('__float__', float),
        ('__oct__', oct), ('__hex__', hex), ('__hash__', hash),
        ('__len__', len), ('__iter__', iter), ('__delattr__', delattr),
        ('__setitem__', setitem), ('__delitem__', delitem),
        ('__setslice__', setslice), ('__delslice__', delslice),
        ('__getitem__', getitem), ('__call__', apply),
        ('__getslice__', getslice), ('__nonzero__', bool)]

    def __init__(klass, name, bases, attributes):
        for k in klass.__magicmethods__:
            if not attributes.has_key(k):
                setattr(klass, k, klass.__forcedmethodname__(k))
        for (k, v) in klass.__magicrmethods__:
            if not attributes.has_key(k):
                setattr(klass, k, klass.__forcedrmethodname__(k, v))
            if not attributes.has_key(v):
                setattr(klass, v, klass.__forcedrmethodname__(v, k))
        for (k, v) in klass.__magicfunctions__:
            if not attributes.has_key(k):
                setattr(klass, k, klass.__forcedmethodfunc__(v))
        super(PromiseMetaClass, klass).__init__(name,
            bases, attributes)

    def __forcedmethodname__(self, method):
        """
        This method builds a forced method. A forced method will
        force all parameters and then call the original method
        on the first argument. The method to use is passed by name.
        """

        def wrapped_method(self, *args):
            result = force(self)
            meth = getattr(result, method)
            args = [force(arg) for arg in args]
            return apply(meth, args)

        return wrapped_method
    
    def __forcedrmethodname__(self, method, alternative):
        """
        This method builds a forced method. A forced method will
        force all parameters and then call the original method
        on the first argument. The method to use is passed by name.
        An alternative method is passed by name that can be used
        when the original method isn't availabe - but with reversed
        arguments. This can only handle binary methods.
        """

        def wrapped_method(self, other):
            self = force(self)
            other = force(other)
            meth = getattr(self, method, None)
            if meth is not None:
                res = meth(other)
                if res is not NotImplemented:
                    return res
            meth = getattr(other, alternative, None)
            if meth is not None:
                res = meth(self)
                if res is not NotImplemented:
                    return res
            return NotImplemented

        return wrapped_method
    
    def __forcedmethodfunc__(self, func):
        """
        This method builds a forced method that uses some other
        function to accomplish it's goals. It forces all parameters
        and then calls the function on those arguments.
        """

        def wrapped_method(*args):
            args = [force(arg) for arg in args]
            return apply(func, args)

        return wrapped_method
    
    def __delayedmethod__(self, func):
        """
        This method builds a delayed method - one that accomplishes
        it's choire by calling some function if itself is forced.
        A class can define a __delayclass__ if it want's to
        override what class is created on delayed functions. The
        default is to create the same class again we are already
        using.
        """

        def wrapped_method(*args, **kw):
            klass = args[0].__class__
            klass = getattr(klass, '__delayclass__', klass)
            return klass(func, args, kw)

        return wrapped_method
    
class Promise(object):

    """
    The initialization get's the function and it's parameters to
    delay. If this is a promise that is created because of a delayed
    method on a promise, args[0] will be another promise of the same
    class as the current promise and func will be one of (getattr,
    apply, getitem, getslice). This knowledge can be used to optimize
    chains of delayed functions. Method access on promises will be
    factored as one getattr promise followed by one apply promise.
    """

    __metaclass__ = PromiseMetaClass

    def __init__(self, func, args, kw):
        """
        Store the object and name of the attribute for later
        resolving.
        """
        self.__func = func
        self.__args = args
        self.__kw = kw
        self.__result = NoneSoFar
    
    def __force__(self):
        """
        This method forces the value to be computed and cached
        for future use. All parameters to the call are forced,
        too.
        """

        if self.__result is NoneSoFar:
            args = [force(arg) for arg in self.__args]
            kw = dict([(k, force(v)) for (k, v)
                    in self.__kw.items()])
            self.__result = apply(self.__func, args, kw)
        return self.__result
    

